import argparse
import os
import sys
from cyvcf2 import VCF, Writer
import pickle
from statistics import mean
import re
from sklearn.ensemble import RandomForestClassifier
import pandas as pd
import numpy as np

def scoring_to_vcf(var,vep_info_index,spliceai_info_index,clf,features,filt):
	#Filt1: Return annotaitons without annotations
	if (var.INFO.get("CSQ") is None
		or var.INFO.get("SpliceAI") is None
	   ) :
		if filt == "off" :
			var.INFO["PDIVAS"] = "None|wo_annots"
			return var
		else :
			return None

   
	# arrange VEP annotations and filtering out annotations outside of Deep intron
	vep_annot = {dict(zip(vep_info_index, x.strip().split("|")))["Gene"]:
				dict(zip(vep_info_index, x.strip().split("|")))
				for x in var.INFO["CSQ"].strip().split(",")
				}
	
	vep_annot = ({k: v for k, v in vep_annot.items()
				  if ('intron_variant' in v["Consequence"]) and 
				  abs(int("".join(re.findall('.+c\.-?\*?[0-9].*([+,-][0-9]+)', v["HGVSc"])) or "0"))>=50})

	# arrange SpliceAI annotations and filtering out annotatins with "." predictions
	spliceai_annot = {dict(zip(spliceai_info_index, x.strip().split("|")))["GENE_ID"]:
					dict(zip(spliceai_info_index, x.strip().split("|")))
					for x in var.INFO["SpliceAI"].strip().split(",")
					}

	spliceai_annot = ({k: v for k, v in spliceai_annot.items() 
						if v["DS_AG"]!="."}) 

	#Filt2: Return annotaitons without candidates
	if (var.CHROM=="Y") | (var.CHROM=="chrY") | (vep_annot == {}) | (spliceai_annot == {}) :
		if filt == "off" :
			var.INFO["PDIVAS"] = "None|out_of_scope"
			return var 
		else :
			return None

	#Get ConSplice and MES_max
	for gene, annot in vep_annot.items():
		#Prepare ConSplice
		vep_annot[gene]["ConSplice"] = (max(annot["ConSplice"].split("&")) if "&" in annot["ConSplice"] 
								  else -1 if annot["ConSplice"]==""
								  else annot["ConSplice"] )
		#Prepare MES
		annot["MES-SWA_acceptor_alt"] = float(annot["MES-SWA_acceptor_alt"] or "0")
		annot["MES-SWA_donor_alt"] = float(annot["MES-SWA_donor_alt"] or "0")
		annot["MES-SWA_acceptor_diff"] = float(annot["MES-SWA_acceptor_diff"] or "0")
		annot["MES-SWA_donor_diff"] = float(annot["MES-SWA_donor_diff"] or "0")
		vep_annot[gene]["MES_SA"] = annot["MES-SWA_acceptor_alt"] if (annot["MES-SWA_acceptor_alt"] > 0 and annot["MES-SWA_acceptor_diff"]) < 0 else 0
		vep_annot[gene]["MES_SD"] = annot["MES-SWA_donor_alt"] if (annot["MES-SWA_donor_alt"] > 0 and annot["MES-SWA_donor_diff"]) < 0 else 0
		vep_annot[gene]["MES_max"] = max([annot["MES_SA"],annot["MES_SD"]])

		del (vep_annot[gene]["MES-SWA_acceptor_diff"], vep_annot[gene]["MES-SWA_donor_diff"], 
			 vep_annot[gene]["MES-SWA_acceptor_alt"], vep_annot[gene]["MES-SWA_donor_alt"])

	#Get SpliceAI_del_gain_max,mean,raw_gain_mean
	for gene, annot in spliceai_annot.items():
		spliceai_annot[gene]["DS_G_MAX"] = max(float(annot["DS_AG"]),float(annot["DS_DG"])) if annot["DS_AG"] != "." else -1
		spliceai_annot[gene]["DS_G_MEAN"] = mean([float(annot["DS_AG"]),float(annot["DS_DG"])]) if annot["DS_AG"] != "." else -1
		spliceai_annot[gene]["RS_G_MEAN"] = mean([float(annot["RS_AG"]),float(annot["RS_DG"])]) if annot["DS_AG"] != "." else -1

	#Gene match between SpliceAI and ConSplice
	combined_score_list = []
	for gene_id in spliceai_annot :
		
		if (gene_id in vep_annot) & (spliceai_annot[gene_id]["DS_G_MAX"]!=-1) :
			combined_score_list.append(
				{"GENE_ID" : gene_id,
				 "AI_del_gain_max" : spliceai_annot[gene_id]["DS_G_MAX"],
				 "AI_del_gain_mean" : spliceai_annot[gene_id]["DS_G_MEAN"],
				 "AI_raw_gain_mean" : spliceai_annot[gene_id]["RS_G_MEAN"],
				 "ConSplice" : vep_annot[gene_id]["ConSplice"],
				 "MES_max" : vep_annot[gene_id]["MES_max"]	
				})

	#Return annotations with matched genes
	if len(combined_score_list)>0 :
		new_anno = []
		var_df = pd.DataFrame(combined_score_list)
		var_df["PDIVAS"] = clf.predict_proba(var_df[features])[:,1]
		new_anno = ["{}|{}".format(d["GENE_ID"], round(d["PDIVAS"],3) \
					if (var.CHROM != "Y") and (var.CHROM != "Mt") and (d["ConSplice"]!=-1) else "out_of_scope") \
					for d in var_df.to_dict("records")]
		var.INFO["PDIVAS"] = ",".join(new_anno)
		return var

	#Filt3: Return annotations without matched genes
	else :
		if filt == "off" :
			var.INFO["PDIVAS"] = "None|no_gene_match"
			return var 
		else :
			return None




import os

__version__ = '1.2.0'
__author__ = 'Ryo Kurosawa'
__license__      = 'MIT'
__author_email__ = 'a0160561@yahoo.co.jp'
__url__          = 'https://github.com/shiro-kur/PDIVAS'

__cur_path__ = __file__




import argparse
import os
import sys
from cyvcf2 import VCF, Writer
import pickle
from statistics import mean
import re
from sklearn.ensemble import RandomForestClassifier
import pandas as pd
import numpy as np
import csv

from .__init__ import __cur_path__, __version__
from pdivas.scoring_to_vcf import scoring_to_vcf
from pdivas.vcf2tsv import vep_editor
from pdivas.vcf2tsv import vep_ai_matcher

def get_options():
    parser = argparse.ArgumentParser(
        prog="pdivas",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
                " ========================================================================================\n"
                "|| PDIVAS: Pathogenicity predictor for Deep-Intronic Variants causing Aberrant Splicing ||\n"
                " ========================================================================================\n"),
                        )
    
    # Calculator of PDIVAS scores
    subparsers = parser.add_subparsers(dest="subcommand")
    
    pred_parser = subparsers.add_parser("predict")
    pred_parser.add_argument(
        "-I",
        required=True,
        metavar="input.vcf/vcf.gz",
        type=str,
        help="The path to the vcf(.gz) file to add PDIVAS annotation")
        
        
    pred_parser.add_argument(
        "-O",
        required=True,
        metavar="output.vcf/vcf.gz",
        type=str,
        help="The path to output vcf(.gz) file name and pass")
        
        
    pred_parser.add_argument(
        "-F",
        metavar="filtering:off/on",
        required=False,
        default="off",
        type=str,
        choices=["off","on"],
        help="Output all variants (-F off; default) or only deep-intronic variants with PDIVAS scores (-F on)")
        
    pred_parser.set_defaults(handler=predict)
    
    
    
    tsv_parser = subparsers.add_parser("vcf2tsv")
    tsv_parser.add_argument(
        "-I",
        required=True,
        metavar="input.vcf/vcf.gz",
        type=str,
        help="The path to the vcf(.gz) file with PDIVAS annotation")
        
        
    tsv_parser.add_argument(
        "-O",
        required=True,
        metavar="output.tsv",
        type=str,
        help="The path to output tsv file name and pass")
        
    tsv_parser.set_defaults(handler=vcf2tsv)
    args = parser.parse_args()
    
    return args,parser
   

def predict(args):
        
    try:
        vcf = VCF(args.I)
    
    except (IOError, ValueError) as e:
        logging.error('{}'.format(e))
        exit()
        
    vep_info_index = vcf["CSQ"]["Description"].strip().split("Format:")[1].strip().replace('"',"").split("|")
    spliceai_info_index = vcf["SpliceAI"]["Description"].strip().split("Format:")[1].strip().replace('"',"").split("|")
    paths = os.path.join(os.path.dirname(str(__cur_path__)),'model/PDIVAS.sav')
    clf = pickle.load(open(paths, 'rb'))
    features = ['AI_del_gain_mean','AI_del_gain_max','AI_raw_gain_mean',"ConSplice","MES_max"]
    
    vcf.add_info_to_header(
        {"ID" :"PDIVAS",
        "Description" : "Predictor of Deep-Intronic Variant causing Aberrant Splicing (PDIVAS). Format: GENE_ID|PDIVAS",
        "Type":"String",
        "Number":"."})

    try:
        oup_file = args.O
        scored_vcf = (
            Writer(oup_file, vcf, "w") if oup_file.endswith("vcf")
            else Writer(oup_file, vcf, "wz") if oup_file.endswith("vcf.gz")
            else Writer(oup_file, vcf, "w")
                    )

    except (IOError, ValueError) as e:
        logging.error('{}'.format(e))
        exit()

    itera = 0
    print("Start calculating PDIVAS")
    for var in vcf:
        itera += 1
        var_oup = scoring_to_vcf(var,vep_info_index,spliceai_info_index,clf,features,args.F)

        if var_oup != None :
            scored_vcf.write_record(var_oup)
        
        else :
            print("scoring error at row",itera,"(",var.CHROM+":"+str(int(var.start)+1)+":"+var.REF+":"+var.ALT[0],")")
            continue
            
    print("PDIVAS successfully executed")
    vcf.close()
    scored_vcf.close()



def vcf2tsv(args):
    
    try:
        vcf = VCF(args.I)
    
    except (IOError, ValueError) as e:
        logging.error('{}'.format(e))
        exit()
    print("Start converting VCF file to TSV file (1 gene annotation per 1 line)")
    
    #Set input/output files
    inp_file = args.I
    vcf = VCF(inp_file)

    fir_var = next(vcf, None)
    fmt_tags = fir_var.FORMAT if fir_var and fir_var.FORMAT else []

    for var in vcf:
        current_tags = var.FORMAT if var.FORMAT else []
        for tag in current_tags:
            if tag not in fmt_tags:
                fmt_tags.append(tag)
    vcf.close()

    vep_info_index = vcf["CSQ"]["Description"].strip().split("Format:")[1].strip().replace('"',"").split("|")
    spliceai_info_index = vcf["SpliceAI"]["Description"].strip().split("Format:")[1].strip().replace('"',"").split("|")
    PD_info_index = ["PD_ID","PDIVAS"]
    SM_index = vcf.samples if vcf.samples else []

    oup_col = ["chr","coordinate","ID","ref","alt","qual","filt"]+vep_info_index+spliceai_info_index+PD_info_index+SM_index
    hgvs_idx = [i for i, x in enumerate(oup_col) if x == 'Gene'][0]

    oup_file = args.O
    csvfile = open(oup_file, "w")
    csv_writer = csv.writer(csvfile, delimiter='\t')
    csv_writer.writerow(["chr","coordinate","ID","ref","alt","qual","filt"]+\
                        vep_info_index+spliceai_info_index+PD_info_index+SM_index)

    itera = 0
    filt_vep_error = 0
    filt_ai_error = 0
    uniq_annots = set([])
    uniq_vars = set([])

    vcf = VCF(inp_file)
    for var in vcf:
        itera += 1
        
        if (var.INFO.get("CSQ") is None) :
            print(var.CHROM, var.start, var.end, var.ID)
            filt_vep_error += 1
            continue

        if (var.INFO.get("SpliceAI") is None) :
            filt_ai_error += 1
            elem = vep_editor(var,vep_info_index,spliceai_info_index,PD_info_index)
        
        else :
            elem = vep_ai_matcher(var,vep_info_index,spliceai_info_index,PD_info_index)
            
        if len(elem) == 0 :
            print("Output error")
            break

        if len(elem) != 0:
            var_ID = ":".join([str(k) for k in [var.CHROM,var.start+1,var.REF,var.ALT[0]]])
            uniq_vars.add(var_ID)

        #Sample columns
        SMs = []
        if len(SM_index)>=1:
            for idx, sample in enumerate(SM_index):
                SM = []
                if "GT" in fmt_tags:
                    GT =str(var.gt_types[idx]) if var.gt_types[idx]!=2 else "NA"
                    GT = "2" if GT =="3" else GT
                    SM.append(GT)
                if "AD" in fmt_tags:
                    AD =str(var.gt_alt_depths[idx])+","+str(var.gt_ref_depths[idx]) if var.gt_alt_depths[idx] != -1 else "NA,NA" 
                    SM.append(AD)
                for tag in [i for i in fmt_tags if i not in ["GT","AD"]] :
                    sm = var.format(tag)[idx] if var.format(tag) is not None else "NA"
                    if isinstance(sm, np.ndarray) :
                        sm = ",".join([str(i) if i > -200000000 else "NA" for i in sm])
                    elif isinstance(sm, np.str_) :
                        sm = sm if sm != "." else "NA"
                    SM.append(sm)
                SMs.append(":".join(SM))

        for gen in range(len(elem)) :
            annot_ID = ":".join([str(k) for k in [var.CHROM,var.start+1,var.REF,var.ALT[0],elem[gen][hgvs_idx]]])
            if annot_ID in uniq_annots :
                print("duplication detected @",annot_ID)
                continue
            uniq_annots.add(annot_ID)
        
            oup_list = elem[gen]+SMs
            csv_writer.writerow(oup_list)


    vcf.close()
    csvfile.close()

    print("# of input variants : ",itera)
    print("# of VEP-lacked variants :",filt_vep_error,";",round(100*filt_vep_error/itera,2),"(%)")
    print("# of AI-lacked variants :",filt_ai_error,";",round(100*filt_ai_error/itera,2),"(%)")
    print("# of output unique variants :",len(uniq_vars),";",round(100*len(uniq_vars)/itera,2),"(%)")
    print("# of output unique annotations : ",len(uniq_annots))

def main(args=None):
    args,parser = get_options()
    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        # show help for unknown subcommands
        parser.print_help()

if __name__ == '__main__':
    main()



from cyvcf2 import VCF, Writer
import csv
import sys

def vep_editor(var,vep_info_index,spliceai_info_index,PD_info_index):
    vep_annot = {dict(zip(vep_info_index, x.strip().split("|")))["Gene"]:
                        x.strip().split("|")
                        for x in var.INFO["CSQ"].strip().split(",")}
    
    base_anno = [var.CHROM,var.start+1,var.ID,var.REF,var.ALT[0],var.QUAL,(var.FILTER or "PASS")]
    oup_list = []
    
    for gene_id in vep_annot:    
        #output formatting
        oup_list += [base_anno+vep_annot[gene_id]+\
                     ["." for i in range(len(spliceai_info_index))]+["wo_annots" for i in range(len(PD_info_index))]]
    
    return oup_list

def vep_ai_matcher(var,vep_info_index,spliceai_info_index,PD_info_index):
    vep_annot = {dict(zip(vep_info_index, x.strip().split("|")))["Gene"]:
                        x.strip().split("|")
                        for x in var.INFO["CSQ"].strip().split(",")}

    spliceai_annot = {dict(zip(spliceai_info_index, x.strip().split("|")))["GENE_ID"]:
                        x.strip().split("|")
                        for x in var.INFO["SpliceAI"].strip().split(",")}

    PD_annot = {dict(zip(PD_info_index, x.strip().split("|")))["PD_ID"]:
                        x.strip().split("|")
                        for x in var.INFO["PDIVAS"].strip().split(",")}

    #Gene match & output formatting
    base_anno = [var.CHROM,var.start+1,var.ID,var.REF,var.ALT[0],var.QUAL,(var.FILTER or "PASS")]
    oup_list = []

    for gene_id in vep_annot:
        if (gene_id in spliceai_annot) :
            #output formatting
            if (gene_id in PD_annot) :
                oup_list += [base_anno+vep_annot[gene_id]+\
                             spliceai_annot[gene_id]+PD_annot[gene_id]]
            else :
                oup_list += [base_anno+vep_annot[gene_id]+\
                             spliceai_annot[gene_id]+list(PD_annot.values())[0]]

        else :
            oup_list += [base_anno+vep_annot[gene_id]+\
                         ["." for i in range(len(spliceai_info_index))]+["wo_annots" for i in range(len(PD_info_index))]]     

    return oup_list



from setuptools import setup, find_packages
from os import path
from codecs import open
import io
import re


#https://www.karakaram.com/how-to-create-python-cli-package/
#https://qiita.com/shinichi-takii/items/6d1063e0aa3f79e599f0

package_name = "pdivas"
root_dir = path.abspath(path.dirname(__file__))

def _requirements():
    return [name.rstrip() for name in open(path.join(root_dir, 'requirements.txt')).readlines()]

with open(path.join(root_dir, package_name, '__init__.py')) as f:
    init_text = f.read()
    version = re.search(r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    license = re.search(r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author = re.search(r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author_email = re.search(r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)

assert version
assert license
assert author
assert author_email
assert url

setup(name='pdivas',
      description='PDIVAS: Pathogenicity predictor of Deep-Intronic Variants causing Aberrant Splicing',
      long_description=io.open('README.md', encoding='utf-8').read(),
      long_description_content_type='text/markdown',
      version=version,
      author=author,
      author_email=author_email,
      license=license,
      url=url,
      packages=find_packages(),
      install_requires=_requirements(),
      package_data={'pdivas': ['model/PDIVAS.sav']},
      entry_points={'console_scripts': ['pdivas=pdivas.__main__:main']}
      )
      
      



